import { getOwner as getOwnerNative, setOwner } from '@ember/application';
import { capabilities } from '@ember/component';
import { setProperties } from '@ember/object';
import { throttle, cancel } from '@ember/runloop';
var HOOKS_STORE = {};
var HOOKS_STATES = {};
var HOOKS_STACK = [];
var BEFORE_CALL_TASKS = [];
var BEFORE_DESTROY_TASKS = [];
export function createHookState(hookName, defaultValue = null) {
    HOOKS_STORE[hookName] = new WeakMap();
    const state = {
        createContext(component) {
            let initValue = typeof defaultValue === 'function' ? defaultValue(component) : [];
            HOOKS_STORE[hookName].set(component, initValue);
        },
        getContext(component) {
            return HOOKS_STORE[hookName].get(component);
        },
        deleteContext(component) {
            return HOOKS_STORE[hookName].delete(component);
        },
        setContext(component, value) {
            return HOOKS_STORE[hookName].set(component, value);
        },
        incrementCallId() {
            this.callId++;
        },
        resetCallId() {
            this.callId = 0;
        },
        callId: 0
    };
    HOOKS_STATES[hookName] = state;
    HOOKS_STACK.push(state);
    return state;
}
function getCurrentContext() {
    return CURRENT_CONTEXT;
}
export function destroyHookState(hookName) {
    HOOKS_STACK = HOOKS_STACK.filter((item) => item !== HOOKS_STATES[hookName]);
    delete HOOKS_STATES[hookName];
}
export function getHookState(hookName) {
    return HOOKS_STATES[hookName];
}
var CURRENT_CONTEXT = null;
var COMPONENTS_COUNTER = 0;
var COMPONENTS_STATES = createHookState('states');
var COMPONENTS_HOOKS = createHookState('hooks');
var COMPONENTS_HOOKS_CACHE_KEYS = createHookState('hooks_cache');
var COMPONENTS_LAYOUT_HOOKS = createHookState('layout_hooks');
var COMPONENTS_LAYOUT_HOOKS_CACHE_KEYS = createHookState('layout_hooks_cache');
var COMPONENTS_TEMPLATES_CONTEXTS = createHookState('templates_context', function () {
    return {};
});
var COMPONENT_HOOKS_CALL_COUNTER = createHookState('call-counter', function () {
    return new WeakMap();
});
function scheduleRerender(compnentContext) {
    compnentContext.renderTimer = throttle(compnentContext, 'update', 16);
}
function compnentSetup(instance) {
    HOOKS_STACK.forEach((hookState) => {
        hookState.createContext(instance);
    });
}
function componentDestroy(instance) {
    beforeRerenderSetup(instance, false);
    BEFORE_DESTROY_TASKS.forEach((task) => {
        task();
    });
    const effects = COMPONENTS_HOOKS.getContext(instance);
    effects.forEach((destroyCb) => {
        if (typeof destroyCb === 'function') {
            destroyCb();
        }
    });
    const layoutEffects = COMPONENTS_LAYOUT_HOOKS.getContext(instance);
    layoutEffects.forEach((destroyCb) => {
        if (typeof destroyCb === 'function') {
            destroyCb();
        }
    });
    cancel(instance.renderTimer);
    HOOKS_STACK.forEach((hookState) => {
        hookState.deleteContext(instance);
    });
    afterRerenderTeardown();
}
export function addBeforeDestroyTask(cb) {
    BEFORE_DESTROY_TASKS.push(cb);
}
export function addBeforeCallTask(cb) {
    BEFORE_CALL_TASKS.push(cb);
}
function beforeRerenderSetup(instance, callHooks = true) {
    CURRENT_CONTEXT = instance;
    if (callHooks) {
        HOOKS_STACK.forEach((hookState) => {
            hookState.resetCallId();
        });
        COMPONENT_HOOKS_CALL_COUNTER.setContext(instance, new WeakMap());
        BEFORE_CALL_TASKS.forEach((task) => {
            task();
        });
    }
}
function afterRerenderTeardown() {
    CURRENT_CONTEXT = null;
}
export function customHook(hookFunction) {
    const currentContext = getCurrentContext();
    const callCounter = COMPONENT_HOOKS_CALL_COUNTER.getContext(currentContext);
    if (currentContext === null) {
        throw new Error('Unable to find component context');
    }
    if (!callCounter.has(hookFunction)) {
        callCounter.set(hookFunction, 0);
    }
    const callId = callCounter.get(hookFunction) || 0;
    return function (...args) {
        const result = hookFunction.apply(null, args);
        callCounter.set(hookFunction, callId + 1);
        return result;
    };
}
export function useState(value) {
    const context = getCurrentContext();
    if (context === null) {
        throw new Error('Unable to find component context');
    }
    const currentStates = COMPONENTS_STATES.getContext(context);
    const callId = COMPONENTS_STATES.callId;
    if (currentStates[callId] === undefined) {
        currentStates.push(value);
    }
    COMPONENTS_STATES.incrementCallId();
    return [currentStates[callId], function (newValue) {
            currentStates[callId] = newValue;
            scheduleRerender(this);
        }.bind(context)];
}
export function getRerender() {
    return scheduleRerender.bind(null, getCurrentContext());
}
export function getService(serviceName) {
    return getOwnerNative(getCurrentContext()).lookup('service:' + serviceName);
}
export function getController(controllerName) {
    return getOwnerNative(getCurrentContext()).lookup('controller:' + controllerName);
}
export function getRoute(routeName) {
    return getOwnerNative(getCurrentContext()).lookup('route:' + routeName);
}
export function getContextId() {
    const context = getCurrentContext();
    if (context === null) {
        throw new Error('Unable to get context id - no context');
    }
    return context.uid;
}
function checkValidationKey(cacheKey, hooksCache, callId) {
    return cacheKey === false || (hooksCache[callId] !== false) && (cacheKey !== hooksCache[callId]);
}
function getValidationKey(cacheKeys) {
    return cacheKeys !== false ? cacheKeys.toString() : false;
}
export function useLayoutEffect(cb, cacheKeys = false) {
    const context = getCurrentContext();
    if (context === null) {
        throw new Error('Unable to find component context');
    }
    const cacheKey = getValidationKey(cacheKeys);
    const currentHooks = COMPONENTS_LAYOUT_HOOKS.getContext(context);
    const hooksCache = COMPONENTS_LAYOUT_HOOKS_CACHE_KEYS.getContext(context);
    const callId = COMPONENTS_LAYOUT_HOOKS.callId;
    if (currentHooks.length - 1 < callId) {
        currentHooks.push(cb);
        hooksCache.push(cacheKey);
    }
    else {
        const mustIvalidateCache = checkValidationKey(cacheKey, hooksCache, callId);
        if (mustIvalidateCache) {
            if (typeof currentHooks[callId] === 'function') {
                let cachedFunction = currentHooks[callId];
                currentHooks[callId] = function () {
                    cachedFunction();
                    cb();
                };
            }
            else {
                currentHooks[callId] = cb;
            }
            hooksCache[callId] = cacheKey;
        }
    }
    COMPONENTS_LAYOUT_HOOKS.incrementCallId();
    return;
}
export function useEffect(cb, cacheKeys = false) {
    const context = getCurrentContext();
    if (context === null) {
        throw new Error('Unable to find component context');
    }
    const cacheKey = getValidationKey(cacheKeys);
    const currentHooks = COMPONENTS_HOOKS.getContext(context);
    const hooksCache = COMPONENTS_HOOKS_CACHE_KEYS.getContext(context);
    const callId = COMPONENTS_HOOKS.callId;
    if (currentHooks.length - 1 < callId) {
        currentHooks.push(cb());
        hooksCache.push(cacheKey);
    }
    else {
        const mustIvalidateCache = checkValidationKey(cacheKey, hooksCache, callId);
        if (mustIvalidateCache) {
            if (typeof currentHooks[callId] === 'function') {
                currentHooks[callId]();
            }
            currentHooks[callId] = cb();
            hooksCache[callId] = cacheKey;
        }
    }
    COMPONENTS_HOOKS.incrementCallId();
    return;
}
function executeLayoutHooks(instance) {
    beforeRerenderSetup(instance);
    const hooks = COMPONENTS_LAYOUT_HOOKS.getContext(instance);
    for (let i = 0; i < hooks.length; i++) {
        if (typeof hooks[i] === 'function') {
            hooks[i] = hooks[i]();
        }
    }
    afterRerenderTeardown();
}
export function getOwner() {
    return getOwnerNative(getCurrentContext());
}
export default class ReactHooksComponentManager {
    static create(attrs) {
        const owner = getOwnerNative(attrs);
        return new this(owner);
    }
    constructor(owner) {
        setOwner(this, owner);
        this.capabilities = capabilities('3.4', {
            destructor: true,
            asyncLifecycleCallbacks: true,
        });
    }
    createCompnentContext(Klass, args = {}) {
        const context = {
            renderFn: Klass,
            uid: COMPONENTS_COUNTER++,
            args,
            renderTimer: null,
            update() {
                beforeRerenderSetup(this);
                setProperties(COMPONENTS_TEMPLATES_CONTEXTS.getContext(this), this.renderFn.call(null, this.args));
                afterRerenderTeardown();
            }
        };
        return context;
    }
    createComponent(Klass, args) {
        let instance = this.createCompnentContext(Klass, args.named);
        setOwner(instance, getOwnerNative(this));
        compnentSetup(instance);
        instance.update();
        return instance;
    }
    updateComponent(instance, args) {
        let currentKeys = Object.keys(instance.args);
        let newKeys = Object.keys(args.named);
        let keys = [].concat(currentKeys, newKeys);
        let unique = [...new Set(keys)];
        let shouldUpdate = false;
        unique.forEach((keyName) => {
            if (instance.args[keyName] !== args.named[keyName]) {
                shouldUpdate = true;
            }
        });
        instance.args = args.named;
        if (shouldUpdate) {
            instance.update();
        }
    }
    destroyComponent(component) {
        componentDestroy(component);
    }
    getContext(component) {
        return COMPONENTS_TEMPLATES_CONTEXTS.getContext(component);
    }
    didCreateComponent(component) {
        executeLayoutHooks(component);
    }
    didUpdateComponent(component) {
        executeLayoutHooks(component);
    }
}
